const sqlite = require('sqlite')
const assert = require('assert')
const path = require('path')
const fs = require('fs')
const wrap = require('co-express')
const PEAnalytics = require('pauls-embedded-analytics')
const {debounceConsole} = require('pauls-log-utils')

const {hashPassword} = require('./crypto')
const figures = require('figures')
const Sessions = require('./sessions')
const Mailer = require('./mailer')
const lock = require('./lock')
const Archiver = require('./archiver')
const UsersAPI = require('./apis/users')
const ArchivesAPI = require('./apis/archives')
const ArchiveFilesAPI = require('./apis/archive-files')
const ReportsAPI = require('./apis/reports')
const ServiceAPI = require('./apis/service')
const UserContentAPI = require('./apis/user-content')
const PagesAPI = require('./apis/pages')
const AdminAPI = require('./apis/admin')
const Shemas = require('./dbs/schemas')
const UsersDB = require('./dbs/users')
const ArchivesDB = require('./dbs/archives')
const ActivityDB = require('./dbs/activity')
const ReportsDB = require('./dbs/reports')
const { isPassword } = require('../lib/validators')

class Hypercloud {
  constructor (config) {
    assert(config, 'hypercloud requires options')
    assert(config.hostname, 'config.hostname is required')
    assert(config.dir || config.db, 'hypercloud requires a dir or db option')
    this.config = config
  }

  async setup () {
    var config = this.config

    // fallback config
    config.env = config.env || 'development'

    // setup config
    var {dir, db} = config
    if (dir) {
      // ensure the target dir exists
      console.log(figures.info, 'Data directory:', dir)
      try {
        fs.accessSync(dir, fs.F_OK)
      } catch (e) {
        fs.mkdirSync(dir)
      }
    }
    if (!db && dir) {
      // allocate a sqlite
      db = await sqlite.open(path.join(dir, 'main.db'))
    }
    assert(db, 'database was not created')
    this.db = db

    // init components
    this.lock = lock
    this.analytics = new PEAnalytics({
      db: path.join(dir, 'analytics.db'),
      domain: config.hostname
    })
    // this.proofs = new Proofs(config) TODO
    this.mailer = new Mailer(config)
    this.archiver = new Archiver(this)
    this.usersDB = new UsersDB(this)
    this.archivesDB = new ArchivesDB(this)
    this.activityDB = new ActivityDB(this)
    this.reportsDB = new ReportsDB(this)
    this.sessions = new Sessions(this)

    // init apis
    this.api = {
      users: new UsersAPI(this),
      archives: new ArchivesAPI(this),
      archiveFiles: new ArchiveFilesAPI(this),
      service: new ServiceAPI(this),
      userContent: new UserContentAPI(this),
      pages: new PagesAPI(this),
      admin: new AdminAPI(this),
      reports: new ReportsAPI(this)
    }

    // wrap all APIs in co-express handling
    wrapAll(this.api.users)
    wrapAll(this.api.archives)
    wrapAll(this.api.archiveFiles)
    wrapAll(this.api.service)
    wrapAll(this.api.userContent)
    wrapAll(this.api.pages)
    wrapAll(this.api.admin)
    wrapAll(this.api.reports)

    // run db migrations
    var schemas = new Shemas(this)
    await schemas.migrate()
    await schemas.runCorrections()

    // final init
    this.archiver.setup()
    await this.usersDB.setup()
    await this.archivesDB.setup()
    await this.activityDB.setup()
    await this.reportsDB.setup()
    await this.setupAdminUser()
  }

  async loadAllArchives () {
    // load all archives
    var ps = []
    var records = await this.archivesDB.listKeys()
    for (let record of records) {
      ps.push(this.archiver.loadArchive(record.key).then(null, _ => null))
      debounceConsole.log(`${figures.pointerSmall} Loading archive`, {timeout: 500, max: 1e3})
    }
    await Promise.all(ps)
    console.log(figures.tick, 'All archives loaded,', ps.length, 'total')

    // compute user disk usage and swarm archives accordingly
    this.archiver.computeAllUserDiskUsageAndSwarm()

    // create the popular-archives index
    this.archiver.computePopularIndex()
  }

  async setupAdminUser () {
    try {
      // is the admin-user config wellformed?
      var adminConfig = this.config.admin
      if (!adminConfig || !adminConfig.password) {
        console.log('Admin user not created: must set password in config')
        return // abort if not
      }
      if (!isPassword(adminConfig.password)) {
        throw new Error('Admin user not created: invalid password')
      }

      // upsert the admin user with these creds
      var method = 'put'
      let {passwordHash, passwordSalt} = await hashPassword(adminConfig.password)
      var adminRecord = await this.usersDB.getByUsername('admin')
      if (!adminRecord) {
        method = 'create'
        adminRecord = {
          username: 'admin',
          scopes: ['user', 'admin:dats', 'admin:users'],
          isEmailVerified: true
        }
      }
      adminRecord.passwordHash = passwordHash
      adminRecord.passwordSalt = passwordSalt
      if (adminConfig.email) adminRecord.email = adminConfig.email
      await this.usersDB[method](adminRecord)
      console.log(figures.tick, (method === 'create' ? 'Created' : 'Updated'), 'admin record')
    } catch (e) {
      console.error('[ERROR] While trying to create admin user:', e)
    }
  }

  async close (cb) {
    await this.archiver.closeAllArchives()
    cb()
  }
}

module.exports = Hypercloud

function wrapAll (api) {
  for (let methodName of Object.getOwnPropertyNames(Object.getPrototypeOf(api))) {
    let method = api[methodName]
    if (typeof method === 'function' && methodName.charAt(0) !== '_') {
      api[methodName] = wrap(method.bind(api))
    }
  }
}
